.App{text-align:left}.navBar{background-color:#fc0}.btnGrey{background-color:#bcbcbb}.btnBlue{background-color:#00a6e1}.btnGreen{background-color:#5dd41b}.btnRed{background-color:#df462e}.textSizeSmallest{font-size:80%!important}.textSizeSmall{font-size:90%!important}.textSizeNormal{font-size:100%!important}.textSizeLarge{font-size:120%!important}.textSizeLargest{font-size:140%!important}.example-enter{opacity:.01}.example-enter.example-enter-active{opacity:1;transition:opacity .5s ease-in}.example-leave{opacity:1}.example-leave.example-leave-active{opacity:.01;transition:opacity .3s ease-in}.example-appear{opacity:.01}.example-appear.example-appear-active{opacity:1;transition:opacity 1s ease-in}.rotate{-webkit-animation:rotation 8s linear infinite;animation:rotation 8s linear infinite}@-webkit-keyframes rotation{0%{transform:rotate(0deg)}to{transform:rotate(359deg)}}@keyframes rotation{0%{transform:rotate(0deg)}to{transform:rotate(359deg)}}
/*# sourceMappingURL=main.45357948.chunk.css.map */